// file      : xsd/cxx/tree/fundamental-header.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#ifndef CXX_TREE_FUNDAMENTAL_HEADER_HXX
#define CXX_TREE_FUNDAMENTAL_HEADER_HXX

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cxx/tree/elements.hxx>

namespace CXX
{
  namespace Tree
  {
    struct FundamentalNamespace : Namespace,

                                  Traversal::AnyType,
                                  Traversal::AnySimpleType,

                                  Traversal::Fundamental::Byte,
                                  Traversal::Fundamental::UnsignedByte,
                                  Traversal::Fundamental::Short,
                                  Traversal::Fundamental::UnsignedShort,
                                  Traversal::Fundamental::Int,
                                  Traversal::Fundamental::UnsignedInt,
                                  Traversal::Fundamental::Long,
                                  Traversal::Fundamental::UnsignedLong,
                                  Traversal::Fundamental::Integer,
                                  Traversal::Fundamental::NonPositiveInteger,
                                  Traversal::Fundamental::NonNegativeInteger,
                                  Traversal::Fundamental::PositiveInteger,
                                  Traversal::Fundamental::NegativeInteger,

                                  Traversal::Fundamental::Boolean,

                                  Traversal::Fundamental::Float,
                                  Traversal::Fundamental::Double,
                                  Traversal::Fundamental::Decimal,

                                  Traversal::Fundamental::String,
                                  Traversal::Fundamental::NormalizedString,
                                  Traversal::Fundamental::Token,
                                  Traversal::Fundamental::Name,
                                  Traversal::Fundamental::NameToken,
                                  Traversal::Fundamental::NameTokens,
                                  Traversal::Fundamental::NCName,
                                  Traversal::Fundamental::Language,

                                  Traversal::Fundamental::Id,
                                  Traversal::Fundamental::IdRef,
                                  Traversal::Fundamental::IdRefs,

                                  Traversal::Fundamental::AnyURI,

                                  Traversal::Fundamental::QName,

                                  Traversal::Fundamental::Base64Binary,
                                  Traversal::Fundamental::HexBinary,

                                  Traversal::Fundamental::Date,
                                  Traversal::Fundamental::DateTime,
                                  Traversal::Fundamental::Duration,
                                  Traversal::Fundamental::Day,
                                  Traversal::Fundamental::Month,
                                  Traversal::Fundamental::MonthDay,
                                  Traversal::Fundamental::Year,
                                  Traversal::Fundamental::YearMonth,
                                  Traversal::Fundamental::Time,

                                  Traversal::Fundamental::Entity,
                                  Traversal::Fundamental::Entities,
                                  protected virtual Context
    {
      FundamentalNamespace (Context& c)
          : Namespace (c, true),
            Context (c)
      {
        *this >> names_ >> *this;
      }

      Void
      built_in_type (SemanticGraph::Type& t,
                     String const& type,
                     String const& name)
      {
        String custom;

        // XML Schema built-in type customization is only possible when
        // we are generating separate header.
        //
        if (options.value<CLI::generate_xml_schema> () &&
            custom_type (t, custom))
        {
          if (custom.empty ())
            custom = name;

          String new_name;
          renamed_type (t, new_name);

          if (new_name)
            os << "typedef " << type << " " << new_name << ";";

          if (custom == name)
            os << "class " << name << ";";
          else
            os << "typedef " << custom << " " << name << ";";
        }
        else
        {
          // Otherwise generate simple typedef.
          //
          os << "typedef " << type << " " << name << ";";
        }
      }

      // anyType and anySimpleType
      //
      virtual Void
      traverse (SemanticGraph::AnyType& t)
      {
        os << "// anyType and anySimpleType." << endl
           << "//" << endl;

        built_in_type (t, "::xsd::cxx::tree::type", "type");
      }

      virtual Void
      traverse (SemanticGraph::AnySimpleType& t)
      {
        built_in_type (
          t, "::xsd::cxx::tree::simple_type<type>", "simple_type");
        os << endl;
      }

      // Integrals.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Byte& t)
      {
        os << "// 8-bit" << endl
           << "//" << endl;

        built_in_type (t, "signed char", "byte");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedByte& t)
      {
        built_in_type (t, "unsigned char", "unsigned_byte");
        os << endl;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Short& t)
      {
        os << "// 16-bit" << endl
           << "//" << endl;

        built_in_type (t, "short", "short_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedShort& t)
      {
        built_in_type (t, "unsigned short", "unsigned_short");
        os << endl;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Int& t)
      {
        os << "// 32-bit" << endl
           << "//" << endl;

        built_in_type (t, "int", "int_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedInt& t)
      {
        built_in_type (t, "unsigned int", "unsigned_int");
        os << endl;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Long& t)
      {
        os << "// 64-bit" << endl
           << "//" << endl;

        built_in_type (t, "long long", "long_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedLong& t)
      {
        built_in_type (t, "unsigned long long", "unsigned_long");
        os << endl;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Integer& t)
      {
        os << "// Supposed to be arbitrary-length integral types." << endl
           << "//" << endl;

        built_in_type (t, "long long", "integer");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonPositiveInteger& t)
      {
        built_in_type (t, "integer", "non_positive_integer");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonNegativeInteger& t)
      {
        built_in_type (t, "integer", "non_negative_integer");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::PositiveInteger& t)
      {
        built_in_type (t, "integer", "positive_integer");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NegativeInteger& t)
      {
        built_in_type (t, "integer", "negative_integer");
        os << endl;
      }

      // Boolean.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Boolean& t)
      {
        os << "// Boolean." << endl
           << "//" << endl;

        built_in_type (t, "bool", "boolean");

        os << endl;
      }

      // Floats.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Float& t)
      {
        os << "// Floating-point types." << endl
           << "//" << endl;

        built_in_type (t, "float", "float_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Double& t)
      {
        built_in_type (t, "double", "double_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Decimal& t)
      {
        built_in_type (t, "long double", "decimal");
        os << endl;
      }


      // Strings.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::String& t)
      {
        os << "// String types." << endl
           << "//" << endl;

        built_in_type (
          t,
          L"::xsd::cxx::tree::string< " + char_type + L", simple_type >",
          "string");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NormalizedString& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::normalized_string< " + char_type +
          L", string >",
          "normalized_string");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Token& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::token< " + char_type + L", normalized_string >",
          "token");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameToken& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::nmtoken< " + char_type + L", token >",
          "nmtoken");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameTokens& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::nmtokens< " + char_type +
          L", simple_type, nmtoken>",
          "nmtokens");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Name& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::name< " + char_type + L", token >",
          "name");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NCName& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::ncname< " + char_type + L", name >",
          "ncname");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Language& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::language< " + char_type + L", token >",
          "language");

        os << endl;
      }

      // ID/IDREF.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Id& t)
      {
        os << "// ID/IDREF." << endl
           << "//" << endl;

        built_in_type (
          t,
          L"::xsd::cxx::tree::id< " + char_type + L", ncname >",
          "id");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRef& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::idref< type, " + char_type + L", ncname >",
          "idref");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRefs& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::idrefs< " + char_type +
          L", simple_type, idref >",
          "idrefs");

        os << endl;
      }


      // URI.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::AnyURI& t)
      {
        os << "// URI." << endl
           << "//" << endl;

        built_in_type (
          t,
          L"::xsd::cxx::tree::uri< " + char_type + L", simple_type >",
          "uri");

        os << endl;
      }

      // Qualified name.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::QName& t)
      {
        os << "// Qualified name." << endl
           << "//" << endl;

        built_in_type (
          t,
          L"::xsd::cxx::tree::qname< " + char_type +
          L", simple_type, uri, ncname >",
          "qname");

        os << endl;
      }

      // Binary.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Base64Binary& t)
      {
        os << "// Binary." << endl
           << "//" << endl
           << "typedef ::xsd::cxx::tree::buffer< " <<
          char_type << " > buffer;";

        built_in_type (
          t,
          L"::xsd::cxx::tree::base64_binary< " + char_type +
          L", simple_type >",
          "base64_binary");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::HexBinary& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::hex_binary< " + char_type +
          L", simple_type >",
          "hex_binary");

        os << endl;
      }


      // Date/time.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Date& t)
      {
        os << "// Date/time." << endl
           << "//" << endl;

        built_in_type (
          t,
          L"::xsd::cxx::tree::date< " + char_type + L", simple_type >",
          "date");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::DateTime& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::date_time< " + char_type + L", simple_type >",
          "date_time");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Duration& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::duration< " + char_type + L", simple_type >",
          "duration");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Day& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::day< " + char_type + L", simple_type >",
          "day");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Month& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::month< " + char_type + L", simple_type >",
          "month");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::MonthDay& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::month_day< " + char_type + L", simple_type >",
          "month_day");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Year& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::year< " + char_type + L", simple_type >",
          "year");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::YearMonth& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::year_month< " + char_type + L", simple_type >",
          "year_month");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Time& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::time< " + char_type + L", simple_type >",
          "time");

        os << endl;
      }

      // Entity.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Entity& t)
      {
        os << "// Entity." << endl
           << "//" << endl;

        built_in_type (
          t,
          L"::xsd::cxx::tree::entity< " + char_type + L", ncname >",
          "entity");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Entities& t)
      {
        built_in_type (
          t,
          L"::xsd::cxx::tree::entities< " + char_type +
          L", simple_type, entity >",
          "entities");

        os << endl;
      }

      virtual Void
      post (SemanticGraph::Namespace& n)
      {
        os << "// Exceptions." << endl
           << "//" << endl
           << "typedef ::xsd::cxx::tree::exception< " <<
          char_type << " > exception;"
           << "typedef ::xsd::cxx::tree::parsing< " <<
          char_type << " > parsing;"
           << "typedef ::xsd::cxx::tree::expected_element< " <<
          char_type << " > expected_element;"
           << "typedef ::xsd::cxx::tree::unexpected_element< " <<
          char_type << " > unexpected_element;"
           << "typedef ::xsd::cxx::tree::expected_attribute< " <<
          char_type << " > expected_attribute;"
           << "typedef ::xsd::cxx::tree::unexpected_enumerator< " <<
          char_type << " > unexpected_enumerator;"
           << "typedef ::xsd::cxx::tree::no_type_info< " <<
          char_type << " > no_type_info;"
           << "typedef ::xsd::cxx::tree::not_derived< " <<
          char_type << " > not_derived;"
           << "typedef ::xsd::cxx::tree::duplicate_id< " <<
          char_type << " > duplicate_id;"
           << "typedef ::xsd::cxx::tree::serialization< " <<
          char_type << " > serialization;"
           << "typedef ::xsd::cxx::tree::no_namespace_mapping< " <<
          char_type << " > no_namespace_mapping;"
           << "typedef ::xsd::cxx::tree::no_prefix_mapping< " <<
          char_type << " > no_prefix_mapping;"
           << "typedef ::xsd::cxx::tree::xsi_already_in_use< " <<
          char_type << " > xsi_already_in_use;"
           << "typedef ::xsd::cxx::tree::bounds< " <<
          char_type << " > bounds;"
           << endl;

        os << "// Parsing/serialization error." << endl
           << "//" << endl
           << "typedef ::xsd::cxx::tree::error< " <<
          char_type << " > error;"
           << "typedef ::xsd::cxx::tree::errors< " <<
          char_type << " > errors;"
           << endl;

        if (!options.value<CLI::suppress_parsing> () ||
            options.value<CLI::generate_serialization> ())
        {
          os << "// Error handler interface." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::xml::error_handler< " <<
            char_type << " > error_handler;"
             << endl;
        }

        if (options.value<CLI::generate_serialization> ())
        {
          os << "// Namespace information. Used in serialization functions." << endl
             << "//" << endl
             << "typedef ::xsd::cxx::xml::dom::namespace_info < " <<
            char_type << " > namespace_info;"
             << "typedef ::xsd::cxx::xml::dom::namespace_infomap < " <<
            char_type << " > namespace_infomap;"
             << endl;
        }

        os << "// Flags and properties." << endl
           << "//" << endl
           << "typedef ::xsd::cxx::tree::flags flags;"
           << "typedef ::xsd::cxx::tree::properties< " << char_type <<
          " > properties;"
           << endl;

        os << "// DOM user data key for back pointers to tree nodes." << endl
           << "//" << endl;

        if (!options.value<CLI::generate_xml_schema> ())
        {
          String g (L"XSD_CXX_TREE_TREE_NODE_KEY_IN_" + ns_name (n));

          std::transform (g.begin (), g.end(), g.begin (), upcase);
          g = escape (g); // Make it a C++ id.

          os << "#ifndef " << g << endl
             << "#define " << g << endl
             << endl;
        }

        os << "const XMLCh* const tree_node_key = " <<
          "::xsd::cxx::tree::user_data_keys::node;"
           << endl;

        if (!options.value<CLI::generate_xml_schema> ())
        {
          os << "#endif" << endl
             << endl;
        }

        Namespace::post (n);
      }

    private:
      Traversal::Names names_;

    };
  }
}

#endif // CXX_TREE_FUNDAMENTAL_HEADER_HXX
